import json
from .smsl_state_branch import smslStateBranch
from .smsl_state import smslState
from .smsl_operation import smslOperation
from .smsl_smsl import post_process

def process_file_data(json_data : dict):
    """
    Process loaded JSON data structure
    """
    sb_list  = []

    # Load the state branches in the json file
    for new_sb_name, new_sb_entry in json_data.items():
        if new_sb_name.startswith('_'):
            # skip non-data entries
            continue
        else:
            initial, activating, num_facts, _sub_sbs = \
                process_file_data_sb_header(new_sb_entry['HEADER'])
            states, operations = \
                process_file_data_sb_states(new_sb_entry)
            if initial == None:
                initial = smslState(states[0])
            if activating == None:
                activating = smslState(states[-1])
            new_sb = smslStateBranch(
                new_sb_name, 
                initial, 
                activating, 
                num_facts, 
                {},
                states,
                operations,
                None,
                _sub_sbs
            )
            sb_list.append(new_sb)

    # Post process the state branches
    sb_list = post_process(sb_list)

    return sb_list

def process_file_data_sb_header(header_data : dict):
    """
    Process the JSON data in the HEADER attribute of an SB
    """
    initial, activating, num_facts, sub_sbs = None, None, 0, {}
    if 'INITIAL' in header_data.keys():
        initial = smslState(header_data['INITIAL'])
    if 'ACTIVATING' in header_data.keys():
        activating = smslState(header_data['ACTIVATING'])
    if 'NUM_FACTS' in header_data.keys():
        num_facts = header_data['NUM_FACTS']
    if 'SUB_SBS' in header_data.keys():
        sub_sbs = header_data['SUB_SBS']
    return initial, activating, num_facts, sub_sbs

def process_file_data_sb_states(sb_data : dict):
    """
    Process the JSON data in the state attributes of an SB
    """
    states, operations = [], []
    for state, ops in sb_data.items():
        if state == "HEADER" or state.startswith('_'):
            # skip non-data entries
            continue
        start_state = smslState(state)
        states.append(start_state)
        operations.extend(
            smslOperation(op, start_state, smslState(end_state)) \
                for op, end_state in ops.items()
        )
    return states, operations

def print_file_data(json_data : dict):
    """
    Print pretty data
    """
    json_formatted_str = json.dumps(json_data, indent=2)
    print(json_formatted_str)

# TODO: Process multiple files 
# (SB can contain sub SBs from other files)
# def process_files_data(json_data : dict):


import networkx as nx
import matplotlib.pyplot as plt
from .smsl_state import smslState
from .smsl_operation import smslOperation
from .smsl_constants import (
    smslDict_SB_Statedigit
)

class smslStateBranch:
    """
    The SB class
    - name
    - initial
    - activating
    - num_facts
    - sub_sbs
    - _sub_sbs
        Temperary dictionary contains the name of sub SBs and
        the corresponding state digit. Is always empty after post-
        processing
    - states
    - operations
    - sb_level
    """
    def __init__(self, 
                 name: str, 
                 initial: smslState = None, 
                 activating: smslState = None, 
                 num_facts: int = 0, 
                 sub_sbs: smslDict_SB_Statedigit = {},
                 states: list[smslState] = [],
                 operations: list[smslOperation] = [],
                 sb_level: int = None,
                 _sub_sbs: dict = {}):
        """
        Constructor
        """
        assert isinstance(initial, smslState), 'Initial state has to be smslState.'
        assert isinstance(activating, smslState), 'Activating state has to be smslState.'
        assert isinstance(num_facts, int), 'Num of facts has to be int.'

        self.name = name
        self.initial = initial
        self.activating = activating
        self.num_facts = num_facts
        self.sub_sbs = sub_sbs
        self._sub_sbs = _sub_sbs
        self.graph = nx.MultiDiGraph()
        self.add_states(states)
        self.add_operations(operations)
        self._post_process()

    def _post_process(self):
        """
        Post processing
        - Get the SB level
        """
        def _assign_levels(node : smslStateBranch, current_level=0):
            node.sb_level = current_level
            for i in node.sub_sbs.keys():
                _assign_levels(i, current_level+1)
        # assign levels
        _assign_levels(self)

    def add_states(self, states : list[smslState]):
        """
        Add states from a list
        """
        self.graph.add_nodes_from(states)
    
    def add_operations(self, operations : list[smslOperation]):
        """
        Add operations from a list
        """
        self.graph.add_edges_from([
            operation._to_graph_edge() for operation in operations
        ])

    def plot_sb(self, with_labels=True):
        """
        Plot the state branch
        """
        # TODO graph plots have issues
        # https://stackoverflow.com/questions/22785849/drawing-multiple-edges-between-two-nodes-with-networkx
        # https://graph-tool.skewed.de
        # TODO plot sub SBs
        pos=nx.spring_layout(self.graph,seed=5)
        nx.draw_networkx_nodes(self.graph, pos)
        nx.draw_networkx_edges(self.graph, pos)
        nx.draw_networkx_labels(self.graph, pos)
        nx.draw_networkx_edge_labels(self.graph, pos)
        
        plt.axis('off')
        axis = plt.gca()
        axis.set_xlim([1.2*x for x in axis.get_xlim()])
        axis.set_ylim([1.2*y for y in axis.get_ylim()])
        plt.tight_layout()
        plt.show()

    def shortest_path(self, start, end):
        """
        Return a list of the states on the shortest path
        """
        return nx.shortest_path(
            self.graph, 
            start, 
            end
        )
    
    def edge_path(self, path : list):
        """
        Return a list of the edges on the shortest path list of states
        """
        return [
            self.graph[path[i]][path[i+1]] \
            for i in range(len(path) - 1)
        ]
    
    def shortest_edge_path(self, start, end):
        """
        Return a list of the edges on the shortest path list of states
        """
        return self.edge_path(self.shortest_path(start, end))


class smslFileNotSupportedError(Exception):
    def __init__(self, message="File format not supported."):
        self.message = message
        super().__init__(self.message)

class smslDataStructNotSupportedError(Exception):
    def __init__(self, message="Data structure not supported."):
        self.message = message
        super().__init__(self.message)

class smslSubSBNotFound(Exception):
    def __init__(self, message="A sub StateBranch was not found."):
        self.message = message
        super().__init__(self.message)

class smslSubSBPostProcessingError(Exception):
    def __init__(self, message="There is an error in the post processing of StateBranch."):
        self.message = message
        super().__init__(self.message)


import copy
from .smsl_state_branch import smslStateBranch
from .smsl_errors import (
    smslSubSBNotFound,
)

def _assign_levels(node : smslStateBranch, current_level=0):
    """
    Helper to assign sb levels recursively
    """
    node.sb_level = current_level
    for i in node.sub_sbs.keys():
        _assign_levels(i, current_level+1)

def post_process(sb_list : list):
    """
    Post processing
    - Attach sub SBs
    - Get the SB level
    - Remove SBs that has an SB level of >= 1 (starting 0)
    """
    # - Attach sub SBs
    # Basically a problem that:
    #   L is a list of tree nodes, randomized order, no connection
    #   Need to reconstruct the tree from these tree nodes
    _temp_list_name_subsb = []
    _temp_name_lookup = {}
    for sb in sb_list:
        _temp_name_lookup[sb.name] = sb
    for sb in sb_list:
        for sub_sb_name, state_digit in sb._sub_sbs.items():
            if sub_sb_name not in _temp_name_lookup.keys():
                raise smslSubSBNotFound
            sb.sub_sbs[copy.deepcopy(_temp_name_lookup[sub_sb_name])] = state_digit
            _temp_list_name_subsb.append(sub_sb_name)
        sb._sub_sbs = {} # clear

    # - Remove SBs that have an SB level of >= 1 (starting 0)
    for sub_sb_name in _temp_list_name_subsb:
        sb_list.remove(_temp_name_lookup[sub_sb_name])

    # - Get the SB level of each SB (starting 0)
    for sb_root in sb_list:
        _assign_levels(sb_root)

    return sb_list


class smslState:
    """
    The State class
    """
    def __init__(self, 
                 name: str):
        """
        Constructor
        """
        self.state_name = name

    def __repr__(self) -> str:
        return f"smslState({self.state_name})"
    
    def __eq__(self, other):
        if isinstance(other, smslState):
            return self.state_name == other.state_name
        return False

    def __hash__(self):
        return hash(self.state_name)
    


from .smsl_state import smslState

class smslOperation:
    """
    The Operation class
    """
    def __init__(self, 
                 name : str,
                 start_state : smslState,
                 end_state : smslState):
        assert isinstance(start_state, smslState), 'Start state has to be smslState.'
        assert isinstance(end_state, smslState), 'End state has to be smslState.'
        self.operation_name = name
        self.start_state = start_state
        self.end_state = end_state

    def _to_graph_edge(self):
        return (self.start_state, self.end_state, {'OP':self.operation_name})
    
    def __repr__(self) -> str:
        return f"smslOperation({repr(self.start_state)} -> {repr(self.end_state)})"


import json, yaml
from pathlib import Path
from .smsl_json import process_file_data as json_process_file_data
from .smsl_yaml import process_file_data as yaml_process_file_data
from .smsl_errors import (
    smslFileNotSupportedError,
    smslDataStructNotSupportedError
)

class smslStateMachine:
    """
    State Machine class
    """
    def __init__(self, file_path=None):
        """
        Initialize State Machine by a file
        """
        if file_path is None:
            self.state_machine = None
            return
            
        file_path = Path(file_path)

        file_data = None
        if file_path.suffix == '.json':
            with open(file_path, 'r') as file:
                file_data = {"TYPE" : "JSON", "DATA" : json.load(file)}
        elif file_path.suffix == '.yaml':
            with open(file_path, 'r') as file:
                file_data = {"TYPE" : "YAML", "DATA" : yaml.safe_load(file)}
        # TODO finish smsl parser
        elif file_path.suffix == '.smsl':
            with open(file_path, 'r') as file:
                # This part might need more work depending on the actual implementation
                from . import smsl_parser
                file_data = {"TYPE" : "SMSL", "DATA" : smsl_parser.load(file)}
        else:
            raise smslFileNotSupportedError(
                f"{file_path.suffix} files are not supported."
            )
        
        self._process_file_data(file_data)
         
    def _process_file_data(self, file_data):
        d_type, d_data = file_data["TYPE"], file_data["DATA"]
        if d_type == "JSON":
            sb_list = json_process_file_data(d_data)
        elif d_type == "YAML":
            sb_list = yaml_process_file_data(d_data)
        # TODO finish smsl parser
        elif file_data["TYPE"] == "SMSL":
            from . import smsl_parser
            sb_list = smsl_parser.process_file_data(file_data["DATA"])
        else:
            raise smslDataStructNotSupportedError(
                f"{d_type} data structure is not supported."
            )
        self.state_machine = sb_list


from setuptools import setup, find_packages

setup(
    name='pysmsl',
    version='0.1.1',
    packages=find_packages(exclude=['smsl.smsl_parser', 'smsl.smsl_parser.*']),
    install_requires=[
        'pyyaml',
        'networkx',
        'matplotlib',
    ],
    author='Yihao Liu',
    author_email='yihao.jhu@gmail.com',
    description='State Machine Serialization Language',
    long_description=open('README.md').read(),
    long_description_content_type='text/markdown',
    url='https://github.com/SMSL-Project/pysmsl',
    license='MIT',
)